//    PCXTranslator 1.0
//    Copyright (C) Marc Schefer, 1999
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <AppKit.h>
#include <translation/TranslatorAddOn.h>

//////////////////////////////////////////////////////////////////////////////
// Misc. definitions
//////////////////////////////////////////////////////////////////////////////

// The following constants are not part of the PCX file format,
// this represents how this program categorizes the PCX files it can read.
typedef enum { PCX_2_COLORS, PCX_16_COLORS, PCX_256_COLORS, PCX_16M_COLORS } PcxType;

const PCX_HEADER_SIZE = 128;
const PCX_PALETTE_SIZE = 769;

const PCX_READ_BUFFER_SIZE = 0x10000; // 64k

const PCX_TYPE = 'PCX ';
#define PCX_MIME "image/x-pcx"
#define PCX_NAME "PCX bitmap format"

const float PCX_INPUT_QUALITY = 0.3; // PCX is certainly not the best format around
const float PCX_INPUT_CAPABILITY = 0.6; // Should be able to decode most PCX files around

//////////////////////////////////////////////////////////////////////////////
// Prototypes
//////////////////////////////////////////////////////////////////////////////

status_t Decode2( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine );
status_t Decode16( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine );
status_t Decode256( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine );
status_t Decode16M( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine );

void DecodeScanline( BPositionIO * inSource, uint8 *inBuffer, int *in, uint8 *decodeBuffer, int pcxBytesPerLine );

//////////////////////////////////////////////////////////////////////////////
// main() does nothing
//////////////////////////////////////////////////////////////////////////////

int main( int argc, char *argv[] )
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Exported data, used by the translation kit
//////////////////////////////////////////////////////////////////////////////

char translatorName[] = "PCXTranslator";
char translatorInfo[] = "PCXTranslator 1.0 Copyright (C) Marc Schefer, 1999";
int32 translatorVersion = 100;	//  Version 1.00

translation_format inputFormats[] =
{
	{ PCX_TYPE, B_TRANSLATOR_BITMAP, PCX_INPUT_QUALITY, PCX_INPUT_CAPABILITY, PCX_MIME, PCX_NAME },
    { 0, 0, 0, 0, 0, 0 } 
};

translation_format outputFormats[] =
{
    { B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 } 
};

//////////////////////////////////////////////////////////////////////////////
// Translation functions
//////////////////////////////////////////////////////////////////////////////

// Called to identify the type of a file if supported
// Identification should be as accurate as possible especially to determine
// whether we'll be able to translate it.

status_t Identify( BPositionIO * inSource,
				const translation_format * inFormat,
				BMessage * ioExtension,
				translator_info * outInfo,
				uint32 outType)
{
	/////////// VERIFY IF OUTPUT TYPE IS SUPPORTED //////////////
	
	if( outType != 0 && outType != B_TRANSLATOR_BITMAP )
		return B_NO_TRANSLATOR; // B_TRANSLATOR_BITMAP is the only supported output format

	/////////// CHECK IF PCX FILE HEADER IS VALID //////////////

	int8 pcxHeader[PCX_HEADER_SIZE];

	if( PCX_HEADER_SIZE != inSource->Read( pcxHeader, PCX_HEADER_SIZE ) )
		return B_NO_TRANSLATOR; // The file is too small to be a valid PCX file or an error occured

	PcxType type;
	// Tests if the file is a valid PCX file and determines its type
	if( pcxHeader[0] == 10 && pcxHeader[1] == 5 && pcxHeader[2] == 1 && pcxHeader[3] == 8 && pcxHeader[65] == 1 )
		type = PCX_256_COLORS;
	else if( pcxHeader[0] == 10 && pcxHeader[1] == 5 && pcxHeader[2] ==1 && pcxHeader[3] == 8 && pcxHeader[65] == 3 )
		type = PCX_16M_COLORS;
	else if( pcxHeader[0] == 10 && pcxHeader[2] ==1 && pcxHeader[3] == 1 && pcxHeader[65] == 4 )
		type = PCX_16_COLORS;
	else if( pcxHeader[0] == 10 && pcxHeader[2] ==1 && pcxHeader[3] == 1 && pcxHeader[65] == 1 )
		type = PCX_2_COLORS;
	else
		return B_NO_TRANSLATOR;

	/////////// GIVE INFO ABOUT THE TRANSLATION //////////////
	
	outInfo->type = PCX_TYPE;
	outInfo->group = B_TRANSLATOR_BITMAP; 
	outInfo->quality = PCX_INPUT_QUALITY; 
	outInfo->capability = PCX_INPUT_CAPABILITY; 
	strcpy( outInfo->name, PCX_NAME ); 
	strcpy( outInfo->MIME, PCX_MIME ); 

    return B_OK;
}

//
// Performs the actual translation work.
//

status_t Translate( BPositionIO * inSource, const translator_info * inInfo,
				    BMessage * ioExtension, uint32 outType, BPositionIO * outDestination )
{
	/////////// VERIFY IF OUTPUT TYPE IS SUPPORTED //////////////
	
	if( outType != 0 && outType != B_TRANSLATOR_BITMAP )
		return B_NO_TRANSLATOR; // B_TRANSLATOR_BITMAP is the only supported output format

	/////////// CHECK IF PCX FILE HEADER IS VALID //////////////

	int8 pcxHeader[PCX_HEADER_SIZE];

	if( PCX_HEADER_SIZE != inSource->Read( pcxHeader, PCX_HEADER_SIZE ) )
		return B_NO_TRANSLATOR; // The file is too small to be a valid PCX file or an error occured

	PcxType type;
	// Tests if the file is a valid PCX file and determines its type
	if( pcxHeader[0] == 10 && pcxHeader[1] == 5 && pcxHeader[2] == 1 && pcxHeader[3] == 8 && pcxHeader[65] == 1 )
		type = PCX_256_COLORS;
	else if( pcxHeader[0] == 10 && pcxHeader[1] == 5 && pcxHeader[2] ==1 && pcxHeader[3] == 8 && pcxHeader[65] == 3 )
		type = PCX_16M_COLORS;
	else if( pcxHeader[0] == 10 && pcxHeader[2] ==1 && pcxHeader[3] == 1 && pcxHeader[65] == 4 )
		type = PCX_16_COLORS;
	else if( pcxHeader[0] == 10 && pcxHeader[2] ==1 && pcxHeader[3] == 1 && pcxHeader[65] == 1 )
		type = PCX_2_COLORS;
	else
		return B_NO_TRANSLATOR;

	/////////// GET RELEVANT HEADER DATA //////////////

	// (converts endianess as apropriate)

	int16 pcxLeft = B_LENDIAN_TO_HOST_INT16( *((int16 *) &pcxHeader[4]) );
	int16 pcxTop = B_LENDIAN_TO_HOST_INT16( *((int16 *) &pcxHeader[6]) );
	int16 pcxRight = B_LENDIAN_TO_HOST_INT16( *((int16 *) &pcxHeader[8]) );
	int16 pcxBottom = B_LENDIAN_TO_HOST_INT16( *((int16 *) &pcxHeader[10]) );
	int16 pcxBytesPerLine = B_LENDIAN_TO_HOST_INT16( *((int16 *) &pcxHeader[66]) );
	
	// Do some tests to catch corrupted files
	if( pcxLeft<0 || pcxRight<0 || pcxTop<0 || pcxBottom<0 ||
	    pcxRight < pcxLeft || pcxBottom < pcxTop )
	    return B_NO_TRANSLATOR;

	/////////// WRITE B_TRANSLATOR_BITMAP HEADER DATA //////////////

	TranslatorBitmap bitmapHeader;
	bitmapHeader.magic = B_TRANSLATOR_BITMAP;
	bitmapHeader.colors = B_RGB32;
	bitmapHeader.bounds.left = pcxLeft;
	bitmapHeader.bounds.top = pcxTop;
	bitmapHeader.bounds.right = pcxRight;
	bitmapHeader.bounds.bottom = pcxBottom;
	bitmapHeader.rowBytes = (pcxRight-pcxLeft+1) * 4;
	bitmapHeader.dataSize = (pcxBottom-pcxTop+1) * bitmapHeader.rowBytes;

	// B_TRANSLATOR_BITMAP is always in BIG ENDIAN format
	swap_data( B_INT32_TYPE, &bitmapHeader.magic, sizeof(int32), B_SWAP_HOST_TO_BENDIAN );
	swap_data( B_INT32_TYPE, &bitmapHeader.colors, sizeof(int32), B_SWAP_HOST_TO_BENDIAN );
	swap_data( B_RECT_TYPE, &bitmapHeader.bounds, sizeof(BRect), B_SWAP_HOST_TO_BENDIAN );
	swap_data( B_INT32_TYPE, &bitmapHeader.rowBytes, sizeof(int32), B_SWAP_HOST_TO_BENDIAN );
	swap_data( B_INT32_TYPE, &bitmapHeader.dataSize, sizeof(int32), B_SWAP_HOST_TO_BENDIAN );
	
	outDestination->Write( &bitmapHeader, sizeof(bitmapHeader) );

	switch( type )
	{
		case PCX_256_COLORS:
			return Decode256( inSource, outDestination, pcxRight-pcxLeft+1, pcxBottom-pcxTop+1, pcxBytesPerLine );
		case PCX_16M_COLORS:
			return Decode16M( inSource, outDestination, pcxRight-pcxLeft+1, pcxBottom-pcxTop+1, pcxBytesPerLine );
		case PCX_16_COLORS:
			return Decode16( inSource, outDestination, pcxRight-pcxLeft+1, pcxBottom-pcxTop+1, pcxBytesPerLine );
		case PCX_2_COLORS:
			return Decode2( inSource, outDestination, pcxRight-pcxLeft+1, pcxBottom-pcxTop+1, pcxBytesPerLine );
	}
}

// decodes the run-length encoding of one complete (all planes) scanline
// (input buffer must be pre-loaded)
void DecodeScanline( BPositionIO * inSource, uint8 *inBuffer, int *in, uint8 *decodeBuffer, int pcxBytesPerLine )
{
	int out=0, count=0;
	while( out < pcxBytesPerLine )
	{
		if(	*in>=PCX_READ_BUFFER_SIZE )
		{
			inSource->Read( inBuffer, PCX_READ_BUFFER_SIZE );
			*in = 0;
		}

		if( count > 0 )
		{
			if( out+count > pcxBytesPerLine )
				break; // This should not happen in valid PCX file, abort
			while( count-- > 0 )
				decodeBuffer[out++] = inBuffer[*in];
		}
		else
		{
			if( 0xc0 == (0xc0 & inBuffer[*in]))
				count = 0x3f & inBuffer[*in];
			else
				decodeBuffer[out++] = inBuffer[*in];
		}
		(*in)++;
	}
}

status_t Decode256( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine )
{
	/////////// READ AND CONVERT PCX PALETTE //////////////
	
	// reads the pcx palette
	uint8 paletteBuffer[PCX_PALETTE_SIZE];
	inSource->Seek( -PCX_PALETTE_SIZE, SEEK_END );
	inSource->Read( paletteBuffer, PCX_PALETTE_SIZE );

	// converts the palette into big endian B_RGB32 format
	uint32 paletteRGB[256];
	for( int i=1, j=0; i<PCX_PALETTE_SIZE; i+=3 )
		paletteRGB[j++] = paletteBuffer[i]<<16 | paletteBuffer[i+1]<<8 | paletteBuffer[i+2];
	
	/////////// READ AND CONVERT PCX IMAGE //////////////

	uint8 *inBuffer = new uint8[PCX_READ_BUFFER_SIZE]; // Allocates a read buffer
	uint8 *decodeBuffer = new uint8[pcxBytesPerLine]; // Allocates a buffer for run-length decoding
	uint32 *outBuffer = new uint32[pcxBytesPerLine];  // Allocates a buffer for format conversion

	// Seek to the beginning of data & Reads a first block of the data
	inSource->Seek( PCX_HEADER_SIZE, SEEK_SET );
	inSource->Read( inBuffer, PCX_READ_BUFFER_SIZE );

	int in = 0;
	for( int i=0; i<pcxHeight; i++ )
	{
		DecodeScanline( inSource, inBuffer, &in, decodeBuffer, pcxBytesPerLine );
		for( int j=0; j<pcxBytesPerLine; j++ )
			outBuffer[j] = paletteRGB[decodeBuffer[j]];

		// writes the bitmap
		outDestination->Write( outBuffer, pcxBytesPerLine * 4 );
	}
	// clean-up
	delete inBuffer;
	delete decodeBuffer;
	delete outBuffer;
	
	return B_OK;
}

status_t Decode16M( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine )
{
	/////////// READ AND CONVERT PCX IMAGE //////////////

	uint8 *inBuffer = new uint8[PCX_READ_BUFFER_SIZE]; // Allocates a read buffer
	uint8 *decodeBuffer = new uint8[pcxBytesPerLine*3]; // Allocates a buffer for run-length decoding
	uint32 *outBuffer = new uint32[pcxBytesPerLine];  // Allocates a buffer for format conversion

	// Seek to the beginning of data & Reads a first block of the data
	inSource->Seek( PCX_HEADER_SIZE, SEEK_SET );
	inSource->Read( inBuffer, PCX_READ_BUFFER_SIZE );

	int in = 0;
	for( int i=0; i<pcxHeight; i++ )
	{
		DecodeScanline( inSource, inBuffer, &in, decodeBuffer, pcxBytesPerLine*3 );

		// Red chanel
		for( int j=0; j<pcxBytesPerLine; j++ )
			outBuffer[j] = decodeBuffer[j] << 16;
		// Green chanel
		for( int j=0; j<pcxBytesPerLine; j++ )
			outBuffer[j] |= decodeBuffer[j+pcxBytesPerLine] << 8;
		// Blue chanel
		for( int j=0; j<pcxBytesPerLine; j++ )
			outBuffer[j] |= decodeBuffer[j+2*pcxBytesPerLine];

		// writes the bitmap
		outDestination->Write( outBuffer, pcxBytesPerLine * 4 );
	}
	
	// clean-up
	delete inBuffer;
	delete decodeBuffer;
	delete outBuffer;
	
	return B_OK;
}

status_t Decode16( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine )
{
	/////////// READ AND CONVERT PCX PALETTE //////////////
	
	// reads the pcx palette
	uint8 paletteBuffer[48];
	inSource->Seek( 16, SEEK_SET );
	inSource->Read( paletteBuffer, 48 );

	// converts the palette into big endian B_RGB32 format
	uint32 paletteRGB[16];
	for( int i=0; i<16; i++ )
		paletteRGB[i] = paletteBuffer[i*3]<<16 | paletteBuffer[i*3+1]<<8 | paletteBuffer[i*3+2];
	
	/////////// READ AND CONVERT PCX IMAGE //////////////

	uint8 *inBuffer = new uint8[PCX_READ_BUFFER_SIZE]; // Allocates a read buffer
	uint8 *decodeBuffer = new uint8[pcxBytesPerLine*4]; // Allocates a buffer for run-length decoding
	uint32 *outBuffer = new uint32[pcxWidth];  // Allocates a buffer for format conversion

	// Seek to the beginning of data & Reads a first block of the data
	inSource->Seek( PCX_HEADER_SIZE, SEEK_SET );
	inSource->Read( inBuffer, PCX_READ_BUFFER_SIZE );

	int in = 0;
	for( int i=0; i<pcxHeight; i++ )
	{
		DecodeScanline( inSource, inBuffer, &in, decodeBuffer, pcxBytesPerLine*4 );

		for( int j=0; j<pcxWidth; j++ )
			outBuffer[j] = (decodeBuffer[j/8] & (1 << (7-j%8))) ? 1 : 0;
		for( int j=0; j<pcxWidth; j++ )
			outBuffer[j] |= (decodeBuffer[j/8 + pcxBytesPerLine] & (1 << (7-j%8))) ? 2 : 0;
		for( int j=0; j<pcxWidth; j++ )
			outBuffer[j] |= (decodeBuffer[j/8 + 2*pcxBytesPerLine] & (1 << (7-j%8))) ? 4 : 0;
		for( int j=0; j<pcxWidth; j++ )
			outBuffer[j] |= (decodeBuffer[j/8 + 3*pcxBytesPerLine] & (1 << (7-j%8))) ? 8 : 0;

		for( int j=0; j<pcxWidth; j++ )
			outBuffer[j] = paletteRGB[outBuffer[j]];

		// writes the bitmap
		outDestination->Write( outBuffer, pcxWidth * 4 );
	}
	
	// clean-up
	delete inBuffer;
	delete decodeBuffer;
	delete outBuffer;
	
	return B_OK;
}

status_t Decode2( BPositionIO * inSource, BPositionIO * outDestination, int pcxWidth, int pcxHeight, int pcxBytesPerLine )
{
	/////////// READ AND CONVERT PCX PALETTE //////////////
	
	// reads the pcx palette
	uint8 paletteBuffer[6];
	inSource->Seek( 16, SEEK_SET );
	inSource->Read( paletteBuffer, 6 );

	// converts the palette into big endian B_RGB32 format
	uint32 paletteRGB[2];
	for( int i=0; i<2; i++ )
		paletteRGB[i] = paletteBuffer[i*3]<<16 | paletteBuffer[i*3+1]<<8 | paletteBuffer[i*3+2];
	
	/////////// READ AND CONVERT PCX IMAGE //////////////

	uint8 *inBuffer = new uint8[PCX_READ_BUFFER_SIZE]; // Allocates a read buffer
	uint8 *decodeBuffer = new uint8[pcxBytesPerLine]; // Allocates a buffer for run-length decoding
	uint32 *outBuffer = new uint32[pcxWidth];  // Allocates a buffer for format conversion

	// Seek to the beginning of data & Reads a first block of the data
	inSource->Seek( PCX_HEADER_SIZE, SEEK_SET );
	inSource->Read( inBuffer, PCX_READ_BUFFER_SIZE );

	int in = 0;
	for( int i=0; i<pcxHeight; i++ )
	{
		DecodeScanline( inSource, inBuffer, &in, decodeBuffer, pcxBytesPerLine );

		for( int j=0; j<pcxWidth; j++ )
			outBuffer[j] = paletteRGB[(decodeBuffer[j/8] & (1 << (7-j%8))) ? 1 : 0];

		// writes the bitmap
		outDestination->Write( outBuffer, pcxWidth * 4 );
	}
	
	// clean-up
	delete inBuffer;
	delete decodeBuffer;
	delete outBuffer;
	
	return B_OK;
}

